# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import logging
import sys
from typing import List, Optional

from databind.json import from_json, to_json
from nr.utils.git import Git  # type: ignore
from termcolor import colored
import click
import yaml

from .. import shut, commons, project
from shut.changelog import v3
from shut.changelog.manager import ChangelogManager
from shut.changelog.render import render as render_changelogs
from shut.model import registry
from shut.model.version import parse_version
from shut.model.package import PackageModel
from shut.utils.cli import editor_open, edit_text


_git = Git()
logger = logging.getLogger(__name__)


@shut.command()
@click.argument('version', type=parse_version, required=False)
@click.option('--reformat', is_flag=True, help='reformat the changelog')
@click.option('--add', metavar='type', help='create a new changelog entry')
@click.option('--for', metavar='component', help='components for the new changelog entry (default: general)', default='general')
@click.option('--fixes', metavar='issue,…', help='issues that this changelog entry fixes')
@click.option('-m', '--message', metavar='text', help='changelog entry description')
@click.option('-e', '--edit', is_flag=True, help='edit the changelog entry or file')
@click.option('--markdown', is_flag=True, help='render the changelog as markdown')
@click.option('-a', '--all', is_flag=True, help='show the changelog for all versions')
@click.option('-s', '--stage', is_flag=True, help='stage the created/updated changelog file with git')
@click.option('-c', '--commit', is_flag=True, help='commit the created/updated changelog file with git, together with other currently staged files')
def changelog(**args):
  """
  Show changelogs or create new entries.
  """

  if (args['version'] or args['reformat']) and args['add']:
    logger.error('unsupported combination of arguments')
    sys.exit(1)

  project.load()
  monorepo = project.monorepo
  package = project.subject if project.subject != monorepo else None

  def _split(s: Optional[str]) -> List[str]:
    return list(filter(bool, map(str.strip, (s or '').split(','))))

  manager = ChangelogManager(
    package.get_changelog_directory() if package
    else monorepo.get_changelog_directory() if monorepo
    else '.changelog')

  if args['add']:

    if not args['for']:
      args['for'] = 'general'

    try:
      type_ = v3.Changelog.Entry.Type[args['add']]
    except KeyError:
      logger.error('invalid changelog type: %r', args['add'])
      sys.exit(1)

    fixes = ['#' + f if f.isdigit() else f for f in _split(args['fixes'])]
    entry = v3.Changelog.Entry(
      type_,
      args['for'],
      args['message'] or '',
      fixes)

    # Allow the user to edit the entry if no description is provided or the
    # -e,--edit option was set.
    if not entry.description or args['edit']:
      serialized = yaml.safe_dump(to_json(entry, v3.Changelog.Entry, registry=registry), sort_keys=False)
      entry = from_json(v3.Changelog.Entry, yaml.safe_load(edit_text(serialized)), registry=registry)

    # Validate the entry contents (need a description and at least one type and component).
    if not entry.description or not entry.component:
      logger.error('changelog entries need a component and description')
      sys.exit(1)

    created = not manager.unreleased.exists()
    manager.unreleased.add_entry(entry)
    manager.unreleased.save(create_directory=True)
    message = ('Created' if created else 'Updated') + ' "{}"'.format(manager.unreleased.filename)
    print(colored(message, 'cyan'))

    if args['stage'] or args['commit']:
      _git.add([manager.unreleased.filename])
    if args['commit']:
      commit_message = entry.description
      if package and monorepo:
        commit_message = '{}({}): '.format(entry.type_.name, package.name) + commit_message
      else:
        commit_message = '{}: '.format(entry.type_.name) + commit_message
      if fixes:
        commit_message += '\n\nfixes ' + ', '.join(fixes)
      _git.commit(commit_message)

    sys.exit(0)

  if args['edit']:
    if not manager.unreleased.exists():
      logger.error('no staged changelog')
      sys.exit(1)
    sys.exit(editor_open(manager.unreleased.filename))

  changelogs = []
  if args['version'] or not args['all']:
    if args['all']:
      sys.exit('error: incompatible arguments: <version> and -a,--all')
    changelog = manager.version(args['version']) if args['version'] else manager.unreleased
    # Load the changelog for the specified version or the current staged entries.
    if not changelog.exists():
      print('No changelog for {}.'.format(colored(str(args['version'] or 'unreleased'), 'yellow')))
      sys.exit(0)
    changelogs.append(changelog)
  else:
    changelogs = list(manager.all())

  if args['reformat']:
    for changelog in changelogs:
      changelog.save()
    sys.exit(0)

  if args['markdown']:
    changelog_format = 'markdown'
  else:
    changelog_format = 'terminal'

  render_changelogs(sys.stdout, changelog_format, changelogs)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import logging
import shlex
import subprocess
import sys
from typing import Dict, List

import click
from nr.stream import concat  # type: ignore
from termcolor import colored

from shut.commands import project
from shut.model import dump, PackageModel
from shut.model.requirements import Requirement, VersionSelector, VendoredRequirement
from . import pkg

logger = logging.getLogger(__name__)


@pkg.group()
def requirements():
  """
  Update requirements in the package configuration.
  """


@requirements.command(no_args_is_help=True)
@click.argument('packages', nargs=-1)
@click.option('-v', '--vendor', 'vendored_packages', multiple=True,
  help='Specify a vendored installation source. This can point to a subdirectory of your project '
  'or a Git repository URL in a format that Pip understands (i.e. starting with git+). This is '
  'useful if a dependency is not available on any package index but only exists in a Git '
  'repository (in which case it can be tracked as a submodule and added with this option).')
@click.option('--test', is_flag=True, help='Add the requirement for testing.')
@click.option('--develop/--no-develop', default=True, help='Specifies whether vendored packages '
  'referencing a local directory should be installed in develop mode or not (default: true)')
def add(packages, test, vendored_packages, develop):
  """
  Install packages and save them to the package requirements.

  Note that this will remove any extranous whitespaces, empty lines and comments
  from the package configuration.
  """

  package = project.load_or_exit(expect=PackageModel)
  if not packages and not vendored_packages:
    sys.exit('error: no packages specified')

  reqs = [Requirement.parse(x) for x in packages]
  vendored_reqs = [VendoredRequirement.parse(x, fallback_to_path=True) for x in vendored_packages]
  target = package.test_requirements if test else package.requirements

  # Ensure that the same vendored requirement doesn't already exist.
  duplicate = next((x for x in vendored_reqs if x in target), None)
  if duplicate:
    logger.warning(f'vendored package {colored("%s", "cyan")} already exists', duplicate)

  # Install the requirements with Pip.
  python = shlex.split(os.getenv('PYTHON', 'python'))
  pip = python + ['-m', 'pip']
  command = pip + ['install'] + [r.to_setuptools() for r in reqs]
  command += concat(x.get_pip_args(package.get_directory(), develop) for x in vendored_reqs)
  command += package.install.get_pip_args()
  res = subprocess.call(command)
  if res != 0:
    sys.exit('error: pip install failed')

  scoped_reqs = [r for r in reqs if r.version != VersionSelector.ANY]
  unscoped_packages = [r.package for r in reqs if r.version == VersionSelector.ANY]

  if unscoped_packages:
    package_versions = get_pip_versions(pip, unscoped_packages)
    scoped_reqs += [Requirement.parse(f'{k} ^{v}') for k, v in package_versions.items()]

  # Update any existing requirements.
  seen = set()
  for req in target.reqs():
    matching_req = next((x for x in scoped_reqs if x.package == req.package), None)
    if matching_req:
      seen.add(req.package)
      req.version = matching_req.version

  # Append the remaining reqs.
  target += [r for r in scoped_reqs if r.package not in seen]
  target += [r for r in vendored_reqs if r not in target]

  dump(package, package.filename)


def get_pip_versions(pip: List[str], package_names: List[str]) -> Dict[str, str]:
  command = pip + ['show'] + package_names
  output = subprocess.check_output(command).decode()
  result = {}
  for chunk in output.split('---'):
    current = {}
    for line in chunk.splitlines():
      key, value = line.partition(':')[::2]
      if key:
        current[key] = value.strip()
    result[current['Name']] = current['Version']
  return result



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import collections
import contextlib
import json
import logging
import os
import posixpath
import re
import textwrap
from typing import Dict, Iterable, List, Optional, TextIO, Tuple

import nr.fs  # type: ignore

from shut.model.package import PackageModel, Include, Exclude
from shut.model.requirements import RequirementsList
from shut.utils.io.virtual import VirtualFiles
from .core import Renderer, register_renderer, VersionRef

logger = logging.getLogger(__name__)

GENERATED_FILE_REMARK = '''
# This file was auto-generated by Shut. DO NOT EDIT
# For more information about Shut, check out https://pypi.org/project/shut/
'''.strip() + '\n'

_ReadmeStatus = collections.namedtuple('ReadmeStatus', 'path,runtime_path,outside')


def _normpath(x):
  return os.path.normpath(x).replace(os.sep, '/')


def _get_readme_content_type(filename: str) -> str:
  return {
    'md': 'text/markdown',
    'rst': 'text/x-rst',
  }.get(nr.fs.getsuffix(filename), 'text/plain')


def _split_section(data, begin_marker, end_marker):
  start = data.find(begin_marker)
  end = data.find(end_marker, start)
  if start >= 0 and end >= 0:
    prefix = data[:start]
    middle = data[start+len(begin_marker):end]
    suffix = data[end+len(end_marker)+1:]
    return prefix, middle, suffix
  return (data, '', '')


@contextlib.contextmanager
def _rewrite_section(fp, data, begin_marker, end_marker):
  """
  Helper to rewrite a section of a file delimited by *begin_marker* and *end_marker*.
  """

  prefix, suffix = _split_section(data, begin_marker, end_marker)[::2]
  fp.write(prefix)
  fp.write(begin_marker + '\n')
  yield fp
  fp.write(end_marker + '\n')
  fp.write(suffix)


class SetuptoolsRenderer(Renderer[PackageModel]):

  #: Begin an end section for the MANIFEST.in file.
  _BEGIN_SECTION = '# This section is auto-generated by Shut. DO NOT EDIT {'
  _END_SECTION = '# }'

  #: These variables are used to format entrypoints in the setup.py file. It
  #: allows the addition of the Python interpreter version to the entrypoint
  #: names.
  _ENTRTYPOINT_VARS = {
    'python-major-version': 'sys.version[0]',
    'python-major-minor-version': 'sys.version[:3]'
  }

  def _render_setup(
    self,
    fp: TextIO,
    package: PackageModel,
  ) -> None:
    metadata = package.get_python_package_metadata()
    install = package.install

    # Write the header/imports.
    fp.write(GENERATED_FILE_REMARK + '\n')
    fp.write('from __future__ import print_function\n')
    if install.hooks.before_install or install.hooks.after_install:
      fp.write('from setuptools.command.install import install as _install_command\n')
    if install.hooks.before_develop or install.hooks.after_develop:
      fp.write('from setuptools.command.develop import develop as _develop_command\n')
    fp.write(textwrap.dedent('''
      import io
      import os
      import setuptools
      import sys
    ''').lstrip())

    # Write hook overrides.
    cmdclass = {}
    if install.hooks.any():
      fp.write('\ninstall_hooks = [\n')
      for hook in package.install_hooks:
        fp.write('  ' + json.dumps(hook.normalize().to_json(), sort_keys=True) + ',\n')
      fp.write(']\n')
      fp.write(textwrap.dedent('''
        def _run_hooks(event):
          import subprocess, shlex, os
          def _shebang(fn):
            with open(fn) as fp:
              line = fp.readline()
              if line.startswith('#'):
                return shlex.split(line[1:].strip())
              return []
          for hook in install_hooks:
            if not hook['event'] or hook['event'] == event:
              command = hook['command']
              if command[0].endswith('.py') or 'python' in _shebang(command[0]):
                command.insert(0, sys.executable)
              env = os.environ.copy()
              env['SHUT_INSTALL_HOOK_EVENT'] = event
              res = subprocess.call(command, env=env)
              if res != 0:
                raise RuntimeError('command {!r} returned exit code {}'.format(command, res))
      '''))
    if install.hooks.after_install or install.hooks.before_install:
      fp.write(textwrap.dedent('''
        class install_command(_install_command):
          def run(self):
            _run_hooks('install')
            super(install_command, self).run()
            _run_hooks('post-install')
      '''))
      cmdclass['install'] = 'install_command'
    if install.hooks.before_develop or install.hooks.after_develop:
      fp.write(textwrap.dedent('''
        class develop_command(_develop_command):
          def run(self):
            _run_hooks('develop')
            super(develop_command, self).run()
            _run_hooks('post-develop')
      '''))
      cmdclass['develop'] = 'develop_command'

    license_file = package.get_license_file(True)
    license_file = os.path.normpath(license_file) if license_file else None
    license_file = os.path.relpath(os.path.abspath(license_file), package.get_directory()) if license_file else None
    if license_file and license_file.startswith(os.pardir + os.sep):
      # We need to copy the license from the monorepo.
      self._render_temp_file_copy_function(fp)
      license_src = os.path.relpath(package.get_license_file(True), package.get_directory())
      license_dst = os.path.basename(license_src)
      fp.write('\n_tempcopy({!r}, {!r})\n'.format(license_src, license_dst))

    readme_file, long_description_expr = self._render_readme_code(fp, package)

    # Write the install requirements.
    fp.write('\n')
    self._render_requirements(fp, 'requirements', package.requirements)

    if package.test_requirements:
      self._render_requirements(fp, 'test_requirements', package.test_requirements)
      tests_require = 'test_requirements'
    else:
      tests_require = '[]'

    if package.extra_requirements:
      fp.write('extras_require = {}\n')
      for key, value in package.extra_requirements.items():
        self._render_requirements(fp, 'extras_require[{!r}]'.format(key), value)
      extras_require = 'extras_require'
    else:
      extras_require = '{}'

    exclude_packages = []
    for pkg in package.exclude:
      exclude_packages.append(pkg)
      exclude_packages.append(pkg + '.*')

    if metadata.is_single_module:
      packages_args = '  py_modules = [{!r}],'.format(package.get_modulename())
    else:
      packages_args = '  packages = setuptools.find_packages({src_directory!r}, {exclude_packages!r}),'.format(
        src_directory=package.source_directory,
        exclude_packages=exclude_packages)

    # Find the requirement on Python itself.
    python_requirement = package.get_python_requirement()
    if python_requirement:
      python_requires_expr = repr(python_requirement.version.to_setuptools() if python_requirement else None)
      if package.is_universal():
        # TODO: We still need to find a good way to convert a Requirement using ORs (|) to setuptools format.
        python_requires_expr = 'None,  # {}'.format(python_requires_expr)
    else:
      python_requires_expr = 'None'

    # TODO: data_files/package_data
    # TODO: py.typed must be included in package_data (or include_package_data=True)
    data_files = '[]'

    # MyPy cannot find PEP-561 compatible packages without zip_safe=False.
    # See https://mypy.readthedocs.io/en/latest/installed_packages.html#making-pep-561-compatible-packages
    zip_safe = not package.typed

    # Write the setup function.
    fp.write(textwrap.dedent('''
      setuptools.setup(
        name = {name!r},
        version = {version!r},
        author = {author_name!r},
        author_email = {author_email!r},
        description = {description!r},
        long_description = {long_description_expr},
        long_description_content_type = {long_description_content_type!r},
        url = {url!r},
        license = {license!r},
      {packages_args}
        package_dir = {{'': {src_directory!r}}},
        include_package_data = {include_package_data!r},
        install_requires = requirements,
        extras_require = {extras_require},
        tests_require = {tests_require},
        python_requires = {python_requires_expr},
        data_files = {data_files},
        entry_points = {entry_points},
        cmdclass = {cmdclass},
        keywords = {keywords!r},
        classifiers = {classifiers!r},
        zip_safe = {zip_safe!r},
    ''').rstrip().format(
      name=package.name,
      version=str(package.get_version()),
      packages_args=packages_args,
      author_name=package.get_author().name if package.get_author() else None,
      author_email=package.get_author().email if package.get_author() else None,
      url=package.get_url(),
      license=package.get_license(),
      description=package.description.replace('\n\n', '%%%%').replace('\n', ' ').replace('%%%%', '\n').strip(),
      long_description_expr=long_description_expr,
      long_description_content_type=_get_readme_content_type(readme_file) if readme_file else None,
      extras_require=extras_require,
      tests_require=tests_require,
      python_requires_expr=python_requires_expr,
      src_directory=package.source_directory,
      include_package_data=True,#package.package_data != [],
      data_files=data_files,
      entry_points=self._render_entrypoints(package.entrypoints),
      cmdclass = '{' + ', '.join('{!r}: {}'.format(k, v) for k, v in cmdclass.items()) + '}',
      keywords = package.keywords,
      classifiers = package.classifiers,
      zip_safe=zip_safe,
    ))

    if package.is_universal():
      fp.write(textwrap.dedent('''
          options = {
            'bdist_wheel': {
              'universal': True,
            },
          },
        )
      '''))
    else:
      fp.write('\n)\n')

  def _render_entrypoints(self, entrypoints: Dict[str, List[str]]) -> None:
    if not entrypoints:
      return '{}'
    lines = ['{']
    for key, value in entrypoints.items():
      lines.append('    {!r}: ['.format(key))
      for item in value:
        item = repr(item)
        args = []
        for varname, expr in self._ENTRTYPOINT_VARS.items():
          varname = '{{' + varname + '}}'
          if varname in item:
            item = item.replace(varname, '{' + str(len(args)) + '}')
            args.append(expr)
        if args:
          item += '.format(' + ', '.join(args) + ')'
        lines.append('      ' + item.strip() + ',')
      lines.append('    ],')
    lines[-1] = lines[-1][:-1]
    lines.append('  }')
    return '\n'.join(lines)

  @staticmethod
  def _format_reqs(reqs: RequirementsList, level: int = 0) -> List[str]:
    indent = '  ' * (level + 1)
    reqs = [r for r in reqs.reqs() if r.package != 'python']
    if not reqs:
      return '[]'
    return '[\n' + ''.join(indent + '{!r},\n'.format(x.to_setuptools()) for x in reqs if x.package != 'python') + ']'

  def _render_requirements(self, fp: TextIO, target: str, requirements: RequirementsList):
    fp.write('{} = {}\n'.format(target, self._format_reqs(requirements)))

  def _get_readme_status(self, package: PackageModel) -> Optional[_ReadmeStatus]:
    """
    Returns some information on the readme for a package. The readme can be located outside
    of the package directory, but that needs to be handled special in various cases.
    """

    readme = package.get_readme_file()
    if not readme:
      return None

    # Make sure the readme is relative (we need it relative either way).
    readme = os.path.relpath(readme, package.get_directory())

    # If the readme file is _not_ inside the package directory, the setup.py will
    # temporarily copy it. The filename at setup time is thus just the readme's
    # base filename.
    is_inside = nr.fs.issub(readme)
    if is_inside:
      readme_relative_path = readme
    else:
      readme_relative_path = os.path.basename(readme)

    return _ReadmeStatus(readme, readme_relative_path, not is_inside)

  def _render_temp_file_copy_function(self, fp: TextIO) -> None:
    """
    Renders a Python function called `_tempcopy()` into *fp*. The function can be used to
    temporarily copy a file for the runtime of the setup file. If the destination file
    already exists, nothing will be done.
    """

    if hasattr(fp, '_SetuptoolsRenderer_tempcopy_rendered'):
      return
    fp._SetuptoolsRenderer_tempcopy_rendered = True

    # TODO(NiklasRosenstein): make sure this gets rendered into the file only once.

    fp.write('\n')
    fp.write(textwrap.dedent('''
      def _tempcopy(src, dst):
        import atexit, shutil
        if not os.path.isfile(dst):
          if not os.path.isfile(src):
            print('warning: source file "{}" for destination "{}" does not exist'.format(src, dst))
            return
          shutil.copyfile(src, dst)
          atexit.register(lambda: os.remove(dst))
    ''').lstrip())
    fp.write('\n')

  def _render_readme_code(self, fp: TextIO, package: PackageModel) -> Tuple[Optional[str], Optional[str]]:
    """
    Renders code for the setup.py file, creating a `long_description` variable. If
    a readme file is present or explicitly specified in *package*, that readme file
    will be read for the setup.

    The readme file may be locatated outside of the packages' directory. In this case,
    the setup.py file will temporarily copy it into the package root directory during
    the setup.

    Returns the Python expression to pass into the `long_description` field of the
    #setuptools.setup() call.
    """

    readme = self._get_readme_status(package)
    if not readme:
      return None, None

    fp.write('\nreadme_file = {!r}\n'.format(readme.runtime_path))

    if readme.outside:
      # Copy the relative README file if it exists.
      self._render_temp_file_copy_function(fp)
      fp.write('_tempcopy({!r}, readme_file)\n'.format(readme.path))

    # Read the contents of the file into the "long_description" variable.
    fp.write(textwrap.dedent('''
      if os.path.isfile(readme_file):
        with io.open(readme_file, encoding='utf8') as fp:
          long_description = fp.read()
      else:
        print("warning: file \\"{}\\" does not exist.".format(readme_file), file=sys.stderr)
        long_description = None
    ''').lstrip())

    return readme.path, 'long_description'

  def _render_manifest_in(self, fp: TextIO, current: TextIO, package: PackageModel) -> None:
    """
    Modifies a `MANIFEST.in` file in place, ensuring that the automatically generatd content
    is up to date (or added if it didn't exist before).
    """

    files = [
      package.filename,
      package.get_py_typed_file(),
      package.get_license_file(True),
    ]

    readme = self._get_readme_status(package)
    if readme:
      files.append(readme.path)

    manifest = [
      os.path.relpath(os.path.abspath(f), package.get_directory())
      for f in files
      if f
    ]

    # Paths outside of the package directory are copied into the package directory
    # on setup.py.
    manifest = [
      os.path.basename(f) if f.startswith(os.pardir + os.sep) else f
      for f in manifest
    ]

    manifest = ['include ' + s for s in manifest]
    metadata = package.get_python_package_metadata()
    for entry in package.package_data:
      if isinstance(entry, Include):
        verb = 'include'
        path = entry.include
      elif isinstance(entry, Exclude):
        verb = 'exclude'
        path = entry.exclude
      else:
        raise RuntimeError(f'unexpected package_data entry: {entry!r}')
      path = posixpath.join(package.source_directory, package.get_modulename().replace('.', '/'), path)
      manifest.append(f'{verb} {path}')

    markers = (self._BEGIN_SECTION, self._END_SECTION)
    with _rewrite_section(fp, current.read() if current else '', *markers):
      for entry in manifest:
        fp.write(entry + '\n')

  # Renderer[PackageModel] Overrides

  def get_files(self, files: VirtualFiles, package: PackageModel) -> None:
    files.add_dynamic('setup.py', self._render_setup, package)
    files.add_dynamic('MANIFEST.in', self._render_manifest_in, package, inplace=True)

    if package.typed:
      directory = package.get_python_package_metadata().package_directory
      files.add_static(os.path.join(directory, 'py.typed'), '')

    if package.has_vendored_requirements():
      logger.info('package "%s" has vendored requirements which will prevent it from '
        'being published.', package.name)

  def get_version_refs(self, package: PackageModel) -> Iterable[VersionRef]:
    def _regex_refs(filename: Optional[str], regex: str) -> Iterable[VersionRef]:
      if filename and os.path.isfile(filename):
        with open(filename) as fp:
          text = fp.read()
          match = re.search(regex, text, re.M)
          if match:
            yield VersionRef(filename, match.start(1), match.end(1), match.group(1))

    filename = os.path.join(package.get_directory(), 'setup.py')
    yield from _regex_refs(filename, r'^\s*version\s*=\s*[\'"]([^\'"]+)[\'"]')

    filename = package.get_python_package_metadata().filename
    yield from _regex_refs(filename, r'^__version__\s*=\s*[\'"]([^\'"]+)[\'"]')


register_renderer(PackageModel, SetuptoolsRenderer)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import abc
from typing import Generic, Iterable, List, T, Type

from nr.stream import concat  # type: ignore

from shut.model import AbstractProjectModel
from shut.model.target import Target, TargetId
from shut.utils.type_registry import TypeRegistry

__all__ = [
  'Publisher',
  'PublisherProvider',
  'register_publisher_provider',
  'get_publishers',
]


class Publisher(Target, metaclass=abc.ABCMeta):

  @abc.abstractmethod
  def get_description(self) -> str:
    pass

  @abc.abstractmethod
  def get_build_dependencies(self) -> Iterable[TargetId]:
    """
    Return the IDs of build targets that this publisher depends on.
    """

  @abc.abstractmethod
  def publish(self, files: List[str], test: bool, verbose: bool) -> bool:
    """
    Run the publishing logic. The builders resolved from #get_build_dependencies() are
    passed to this function. They will already be built when this method is called.
    """


class PublisherProvider(Generic[T], metaclass=abc.ABCMeta):

  @abc.abstractmethod
  def get_publishers(self, obj: T) -> Iterable[Publisher]:
    """
    Return the publishers provided by this plugin.
    """


registry = TypeRegistry[PublisherProvider[AbstractProjectModel]]()


def register_publisher_provider(type_: Type[T], provider_class: Type[PublisherProvider[T]]) -> None:
  registry.put(type_, provider_class)


def get_publishers(obj: T) -> Iterable[Publisher]:
  return concat(provider().get_publishers(obj) for provider in registry.for_type(type(obj)))



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import subprocess as sp
import sys

import click
from termcolor import colored

from shut.commands import shut
from shut.model.package import PackageModel
from shut.test import Runtime, TestRun, TestStatus, Virtualenv
from shut.utils.text import indent_text
from . import pkg, project


def test_package(
  package: PackageModel,
  isolate: bool,
  keep_test_env: bool = False,
  capture: bool = True,
) -> TestRun:
  if not package.test_driver:
    raise RuntimeError('package has no test driver configured')
  if isolate:
    venv = Virtualenv(os.path.join(package.get_directory(), '.venv-test'))
    if venv.exists():
      print('Re-using existing virtual environment at .venv-test ...')
    else:
      print('Creating temporary virtual environment at .venv-test ...')
      venv.create(Runtime.from_env())
    runtime = venv.get_runtime()
    print(f'Installing package "{package.name}" and test requirements ...')
    try:
      orig_cwd = os.getcwd()
      os.chdir(package.get_directory())
      shut(['pkg', '--no-checks', 'install', '--pip', venv.bin('pip'), '--extra', 'test', '-q'], standalone_mode=False)
    except SystemExit as exc:
      os.chdir(orig_cwd)
      if exc.code != 0:
        raise
  else:
    venv = None
    runtime = Runtime.from_env()

  test_reqs = [req.to_setuptools() for req in package.test_driver.get_test_requirements()]
  if test_reqs:
    sp.check_call(runtime.pip + ['install', '-q'] + test_reqs)

  try:
    return package.test_driver.test_package(package, runtime, capture)
  finally:
    if venv and not keep_test_env:
      venv.rm()


def print_test_run(test_run: TestRun) -> None:
  if test_run.status == TestStatus.ERROR:
    print(f'There was an unexpected error when running the tests.')
    if test_run.error:
      print()
      print(colored(indent_text(test_run.error.strip(), 4), 'red'))
      print()
    return

  sorted_tests = sorted(test_run.tests, key=lambda t: t.name)

  # Print a summary.
  n_passed = sum(1 for t in test_run.tests if t.status == TestStatus.PASSED)
  n_skipped = sum(1 for t in test_run.tests if t.status == TestStatus.SKIPPED)
  status_line = (
    f'Ran {len(test_run.tests)} test(s) in {test_run.duration:.3f}s '
    f'({n_passed} passed, {n_skipped} skipped, {len(test_run.tests) - n_passed - n_skipped} failed, '
    f'{len(test_run.errors)} error(s)). {test_run.status.name}')
  print(status_line)
  print()
  for test in sorted_tests:
    color = {TestStatus.PASSED: 'green', TestStatus.SKIPPED: 'yellow', TestStatus.FAILED: 'red'}[test.status]
    print(f'  {colored(test.name, color, attrs=["bold"])} {test.status.name}')
  if sorted_tests:
    print()

  if n_passed < len(test_run.tests):
    # Print error details.
    print('Failed test details:')
    print('====================')
    not_passed_not_skipped = lambda t: t.status not in (TestStatus.PASSED, TestStatus.SKIPPED)
    for i, test in enumerate(filter(not_passed_not_skipped, sorted_tests)):
      print()
      line = f'  {colored(test.name, "red", attrs=["bold"])} ({test.filename}:{test.lineno})'
      print(line)
      print('  ' + '-' * (len(line) - 2))
      print()
      print(indent_text(test.crash.longrepr, 6))
      if test.stdout:
        print('\n  captured stdout:\n')
        print(indent_text(test.stdout, 6))
    print()

  if test_run.errors:
    header = f'Encountered {len(test_run.errors)} error(s)'
    print(header)
    print('=' * len(header))
    for err in test_run.errors:
      print()
      if err.filename:
        print(f'  {colored(err.filename, "red")}')
        print('  ' + '-' * len(err.filename))
        print()
      else:
        print('  ----')
      print(indent_text(err.longrepr, 6))
    print()

  print(colored(status_line, 'grey'))


@pkg.command()
@click.option('--isolate/--no-isolate', default=False,
  help='Isolate all test runs in virtual environments. This greatly increases the duration '
       'for tests to run as the environment needs to be created and packages installed first, '
       'but it ensures that the unit tests work for a vanilla installation (default: false)')
@click.option('--keep-test-env', is_flag=True,
  help='Do not delete the virtual environment created when testing with --isolate.')
@click.option('--capture/--no-capture', default=True,
  help='Capture the output of the underlying testing framework. If set to false, the output '
       'will be routed to stderr (default: true)')
def test(isolate: bool, keep_test_env: bool, capture: bool) -> None:
  """
  Run the package's unit tests.
  """

  package = project.load_or_exit(expect=PackageModel)
  test_run = test_package(package, isolate, keep_test_env, capture)
  print_test_run(test_run)
  sys.exit(0 if test_run.status == TestStatus.PASSED else 1)




import os
import typing as t


def get_file_in_directory(directory: str, prefix: str, preferred: t.List[str]) -> t.Optional[str]:
  """
  Returns a file in *directory* that is either in the *preferred* list or starts with
  specified *prefix*.
  """

  choices = []
  for name in sorted(os.listdir(directory)):
    if name in preferred:
      break
    if name.startswith(prefix):
      choices.append(name)
  else:
    if choices:
      return choices[0]
    return None

  return os.path.join(directory, name)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
List or search package classifiers on PyPI.
"""

import click
from shut.utils.external.classifiers import get_classifiers
from . import shut


@shut.group(help=__doc__)
def classifiers():
  pass


@classifiers.command()
def ls():
  " List all classifiers. "

  for classifier in get_classifiers():
    print(classifier)


@classifiers.command()
@click.option('-q', '--term', help='The substring that will be used to filter classifiers.')
def search(term):
  " Search for package classifiers. "

  for classifier in get_classifiers():
    if not term or term.strip().lower() in classifier.lower():
      print(classifier)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
This package implements the Shut CLI.
"""

from shut import __version__
from shut.model import Project
from nr.proxy import proxy

import click
import logging
import os
import warnings
import sys

context: click.Context = proxy(lambda: click.get_current_context().obj)
project: Project = proxy(lambda: click.get_current_context().obj['project'])


@click.group()
@click.option('-C', '--cwd', metavar='path', help='Run as if shut was started inside the specified directory.')
@click.option('-v', '--verbose', count=True, help='Increase the log verbosity.')
@click.option('-q', '--quiet', is_flag=True, help='Quiet mode, wins over --verbose.')
@click.option('-W', '--disable-warnings', is_flag=True, help='Disable Python warnings.')
@click.version_option(__version__)
def shut(cwd, verbose, quiet, disable_warnings):
  """
  Shut is a tool to manage the lifecycle of pure Python packages. It automates tasks such
  as bootstrapping a project, bumping version numbers, managing changelogs and publishing
  packages to PyPI all the while performing sanity checks.

  Shut makes strong assumptions on the project structure and assumes that the source-control
  system of choice is Git.
  """

  if cwd:
    os.chdir(cwd)

  ctx = click.get_current_context()
  ctx.ensure_object(dict)
  context['quiet'] = quiet
  context['project'] = Project()

  if quiet:
    level = logging.CRITICAL
  elif verbose >= 2:
    level = logging.DEBUG
  elif verbose >= 1:
    level = logging.INFO
  else:
    level = logging.WARNING

  logging.basicConfig(
    format='[%(levelname)s|%(asctime)s|%(name)s]: %(message)s',
    level=level,
  )

  if not sys.warnoptions and not disable_warnings:
    warnings.simplefilter("default") # Change the filter in this process


from . import changelog
from . import classifiers
from . import conda_forge
from . import license
from . import mono
from . import pkg



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import logging
import sys

from termcolor import colored

from shut.commands.commons.bump import make_bump_command, VersionBumpData
from shut.model import PackageModel
from shut.model.version import get_commit_distance_version, Version
from shut.renderers import get_files
from shut.utils.io.virtual import VirtualFiles
from . import pkg
from .update import update_package

logger = logging.getLogger(__name__)


class PackageBumpData(VersionBumpData[PackageModel]):

  def loaded(self) -> None:
    error_s = colored('error', 'red', attrs=['bold', 'underline'])

    # Cannot bump a package if the managed files are out-dated.
    files = get_files(self.obj)
    if files.get_modified_files(self.obj.get_directory()):
      command = colored("shut pkg update", "green", attrs=["bold", "underline"])
      print(f'{error_s}: cannot bump package version while some managed files '
            f'are outdated. run  {command} to fix.', file=sys.stderr)
      sys.exit(1)

    # Cannot bump a package that is part of a mono repository and that mono
    # repository has $.release.single-version enabled.
    if self.project.monorepo and self.project.monorepo.release.single_version:
      if self.args.force:
        logger.warning(
          'forcing version bump on individual package version that is usually managed '
          'by the monorepo.')
        return
      print(f'{error_s}: cannot bump package tied to a monorepo single-version.', file=sys.stderr)
      sys.exit(1)

  def update(self, new_version: Version) -> VirtualFiles:
    self.obj.version = new_version
    vfiles = update_package(self.obj, dry=self.args.dry, indent=1)
    return vfiles

  def get_snapshot_version(self) -> Version:
    project = self.project
    if project.monorepo and project.monorepo.release.single_version:
      subject = project.monorepo
    else:
      subject = self.obj
    return get_commit_distance_version(
      subject.directory,
      subject.version,
      subject.get_tag(subject.get_version())) or subject.get_version()


pkg.command()(make_bump_command(PackageBumpData, PackageModel))



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import sys

import click
from nr.stream import groupby, Stream  # type: ignore
from termcolor import colored

from shut.builders import get_builders
from shut.model import PackageModel
from shut.model.target import TargetId
from shut.publishers import get_publishers
from . import pkg
from .build import run_builds
from .. import project


class PublishError(Exception):
  pass


def publish_package(
  package: PackageModel,
  target: TargetId,
  build_dir: str,
  skip_build: bool,
  test: bool,
  verbose: bool,
) -> None:

  if package.has_vendored_requirements():
    raise PublishError('package has vendored requirements and cannot be published')

  publishers = list(get_publishers(package))
  publishers = [p for p in publishers if target.match(p.id)]
  if not publishers:
    raise PublishError(f'no target matches "{target}"')

  # Prepare the builds that need to be built for the publishers.
  all_builders = list(get_builders(package))
  builders_for_publisher = {}
  for publisher in publishers:
    builders = []
    for target_id in publisher.get_build_dependencies():
      matched_builders = [b for b in all_builders if target_id.match(b.id)]
      if not matched_builders:
        raise PublishError(f'publisher "{publisher.id}" depends on build target "{target_id}" '
                           f'which could not be resolved.')
      builders.extend(b for b in matched_builders if b not in builders)
    builders_for_publisher[publisher.id] = builders

  # Build all builders that are needed.
  if not skip_build:
    built = set()
    for publisher in publishers:
      print()
      builders = builders_for_publisher[publisher.id]
      success = run_builds([b for b in builders if b not in built], build_dir, verbose)
      if not success:
        raise PublishError('build step failed')

  # Execute the publishers.
  for publisher in publishers:
    print()
    print(f'publishing {colored(publisher.id, "cyan")}')
    builders = builders_for_publisher[publisher.id]
    files = (Stream
      .concat(b.get_outputs() for b in builders)
      .map(lambda x: os.path.join(build_dir, x))
      .collect()
    )
    for filename in files:
      print(f'  :: {filename}')
    print()
    success = publisher.publish(files, test, verbose)
    if not success:
      raise PublishError('publish step failed')


@pkg.command()
@click.argument('target', type=lambda s: TargetId.parse(s, True), required=False)
@click.option('-t', '--test', is_flag=True, help='publish to the test repository instead')
@click.option('-l', '--list', 'list_', is_flag=True, help='list available publishers')
@click.option('-v', '--verbose', is_flag=True, help='show more output')
@click.option('-b', '--build-dir', default='build', help='build output directory')
@click.option('--skip-build', is_flag=True, help='do not build artifacts that are to be published')
def publish(target, test, list_, verbose, build_dir, skip_build):
  """
  Publish the package to PyPI or another target.
  """

  if list_ and target:
    sys.exit('error: conflicting options')

  package = project.load_or_exit(expect=PackageModel)

  if list_:
    publishers = list(get_publishers(package))
    for scope, publishers in groupby(publishers, lambda p: p.id.scope):
      print(f'{colored(scope, "green")}:')
      for publisher in publishers:
        print(f'  {publisher.id.name} – {publisher.get_description()}')
    return

  if not target:
    sys.exit('error: no target specified')

  if package.has_vendored_requirements():
    sys.exit(f'error: package has vendored requirements and cannot be published')

  if list_:
    for scope, publishers in groupby(publishers, lambda p: p.id.scope):
      print(f'{colored(scope, "green")}:')
      for publisher in publishers:
        print(f'  {publisher.id.name} – {publisher.get_description()}')
    return

  if not target:
    sys.exit('error: no target specified')

  try:
    publish_package(package, target, build_dir, skip_build, test, verbose)
  except PublishError as exc:
    sys.exit(f'error: {exc}')



# This file was auto-generated by Shut. DO NOT EDIT
# For more information about Shut, check out https://pypi.org/project/shut/

from __future__ import print_function
import io
import os
import setuptools
import sys

readme_file = 'README.md'
if os.path.isfile(readme_file):
  with io.open(readme_file, encoding='utf8') as fp:
    long_description = fp.read()
else:
  print("warning: file \"{}\" does not exist.".format(readme_file), file=sys.stderr)
  long_description = None

requirements = [
  'beautifulsoup4 >=4.8.1,<5.0.0',
  'databind.core >=0.4.0,<1.0.0',
  'databind.json >=0.4.0,<1.0.0',
  'click >=7.0.0,<8.0.0',
  'jinja2 >=2.11.1,<3.0.0',
  'networkx >=2.4.0,<3.0.0',
  'nr.fs >=1.5.0,<2.0.0',
  'nr.parsing.date >=0.4.0,<1.0.0',
  'nr.proxy >=1.0.0,<2.0.0',
  'nr.pylang.utils >=0.0.1,<1.0.0',
  'nr.utils.git >=0.1.3,<0.2.0',
  'nr.stream >=0.0.4,<0.1.0',
  'requests >=2.22.0,<3.0.0',
  'packaging >=20.1.0,<21.0.0',
  'PyYAML >=5.1.0,<6.0.0',
  'termcolor >=1.1.0,<2.0.0',
  'twine',
  'wheel',
]
test_requirements = [
  'pytest',
]

setuptools.setup(
  name = 'shut',
  version = '0.11.5',
  author = 'Niklas Rosenstein',
  author_email = 'rosensteinniklas@gmail.com',
  description = 'Automates the heavy lifting of release and distribution management for pure Python packages.',
  long_description = long_description,
  long_description_content_type = 'text/markdown',
  url = 'https://github.com/NiklasRosenstein/shut',
  license = 'MIT',
  packages = setuptools.find_packages('src', ['test', 'test.*', 'tests', 'tests.*', 'docs', 'docs.*']),
  package_dir = {'': 'src'},
  include_package_data = True,
  install_requires = requirements,
  extras_require = {},
  tests_require = test_requirements,
  python_requires = '>=3.6.0,<4.0.0',
  data_files = [],
  entry_points = {
    'console_scripts': [
      'shut = shut.commands:shut',
    ]
  },
  cmdclass = {},
  keywords = [],
  classifiers = [],
  zip_safe = False,
)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import sys
from typing import Any, Dict, List, Optional, TextIO, Tuple, Type, TypeVar, Union

import nr.fs  # type: ignore
import yaml
from databind.core import ConversionError, datamodel, field, Registry
from databind.json import from_json, to_json, registry as json_registry

registry = Registry(json_registry)
registry.set_option(datamodel, 'skip_defaults', True)
T = TypeVar('T')
ExcInfo = Tuple


def get_existing_file(directory: str, choices: List[str]) -> bool:
  for fn in choices:
    path = os.path.join(directory, fn)
    if os.path.isfile(path):
      return path
  return None


class Unexpected(Exception):

  def __init__(self, expected, got):
    self.expected = expected
    self.got = got

  def __str__(self):
    return f'expected: {self.expected}, got {self.got}'


class Project:
  """
  Loads package and mono repo configuration files and caches them to ensure that
  the same filename is never loaded into a different model object.
  """

  monorepo_filenames = ['monorepo.yml', 'monorepo.yaml']
  package_filenames = ['package.yml', 'package.yaml']

  Unexpected = Unexpected

  def __init__(self):
    self._cache: Dict[str, 'AbstractProjectModel'] = {}
    self.subject: 'AbstractProjectModel' = None
    self.monorepo: MonorepoModel = None
    self.packages: List[PackageModel] = []
    self.invalid_packages: List[Tuple[str, ExcInfo]] = []

  def load(
    self,
    directory: str = '.',
    expect: Optional[Type['AbstractProjectModel']] = None,
  ) -> 'AbstractProjectModel':
    """
    Loads all project information from *directory*. This searches in all parent directories
    for a package or monorepo configuration, then loads all resources that belong to the
    project.
    """

    directory = os.path.abspath(directory)
    monorepo_fn = None
    package_fn = None

    # TODO(NiklasRosenstein): Iterate parent dirs until match is found.
    for dirname in [directory]:
      package_fn = get_existing_file(dirname, self.package_filenames)
      if package_fn:
        break
      monorepo_fn = get_existing_file(dirname, self.monorepo_filenames)
      if monorepo_fn:
        break

    if package_fn:
      monorepo_fn = get_existing_file(os.path.dirname(os.path.dirname(package_fn)),
                                      self.monorepo_filenames)

    if monorepo_fn:
      self.subject = self._load_monorepo(monorepo_fn)
    if package_fn:
      self.subject = self._load_package(package_fn)

    if expect and not isinstance(self.subject, expect):
      raise Unexpected(expect, type(self.subject))

    return self.subject

  def reload(self) -> None:
    """
    Reloads the monorepo and packages.
    """

    if self.monorepo:
      self._reload(self.monorepo)

    for package in self.packages:
      self._reload(package)

  def _reload(self, obj: 'AbstractProjectModel') -> None:
    result = self._load_object(obj.filename, type(obj), force=True)
    vars(obj).update(vars(result))

  def load_or_exit(self, *args, **kwargs):
    try:
      return self.load(*args, **kwargs)
    except Unexpected as exc:
      if exc.expected == MonorepoModel:
        sys.exit('error: not in a mono repository context')
      elif exc.expected == PackageModel:
        sys.exit('error: not in a package context')
      else:
        raise

  def _load_object(self, filename: str, type_: Type[T], force: bool = False) -> T:
    filename = os.path.normpath(os.path.abspath(filename))
    if not force and filename in self._cache:
      obj = self._cache[filename]
      assert isinstance(obj, type_), 'type mismatch: have {} but expected {}'.format(
        type(obj).__name__, type_.__name__)
      return obj
    with open(filename) as fp:
      data = yaml.safe_load(fp)
    #node_collector = NodeCollector()
    obj = self._cache[filename] = from_json(type_, data, registry=registry)
    obj.filename = filename
    obj.project = self
    #obj.unknown_keys = list(Stream.concat(
    #    (x.locator.append(k) for k in x.unknowns)
    #    for x in node_collector.nodes))
    return obj

  def _load_monorepo(self, filename: str) -> 'MonorepoModel':
    self.monorepo = self._load_object(filename, MonorepoModel)

    # Load packages in that monorepo.
    directory = os.path.dirname(filename)
    for item_name in os.listdir(directory):
      package_fn = get_existing_file(os.path.join(directory, item_name), self.package_filenames)
      if package_fn:
        try:
          self._load_package(package_fn)
        except ConversionError as exc:
          self.invalid_packages.append((item_name, sys.exc_info()))

    return self.monorepo

  def _load_package(self, filename: str) -> 'PackageModel':
    package = self._load_object(filename, PackageModel)
    if package not in self.packages:
      self.packages.append(package)
    return package


def dump(obj: Any, file_: Union[str, TextIO]) -> None:
  if isinstance(file_, str):
    with nr.fs.atomic_file(file_, 'w') as fp:
      dump(obj, fp)
  else:
    data = to_json(obj, registry=registry)
    yaml.safe_dump(data, file_, sort_keys=False)


from .abstract import AbstractProjectModel
from .monorepo import MonorepoModel
from .package import PackageModel



import click
import networkx as nx

from shut.commands.pkg.install import collect_requirement_args, run_install, split_extras
from shut.model.monorepo import MonorepoModel
from . import mono, project


@mono.command()
@click.option('--develop/--no-develop', default=True,
  help='Install in develop mode (default: true)')
@click.option('--extra', type=split_extras, help='Specify one or more extras to install.')
@click.option('-U', '--upgrade', is_flag=True, help='Upgrade all packages (forwarded to pip install).')
@click.option('-q', '--quiet', is_flag=True, help='Quiet install')
@click.option('--pip', help='Override the command to run Pip. Defaults to "python -m pip" or the PIP variable.')
@click.option('--pip-args', help='Additional arguments to pass to Pip.')
@click.option('--dry', is_flag=True, help='Print the Pip command to stdout instead of running it.')
def install(develop, extra, upgrade, quiet, pip, pip_args, dry):
  """
  Install all packages in the monorepo using`python -m pip`.

  The command used to invoke Pip can be overwritten using the `PIP` environment variable.
  """

  monorepo = project.load_or_exit(expect=MonorepoModel)
  graph = monorepo.get_inter_dependencies_graph()
  package_map = {p.name: p for p in project.packages}

  args = []
  for package_name in nx.algorithms.dag.topological_sort(graph):
    package = package_map[package_name]
    args += collect_requirement_args(package, develop, False, extra)
    args += package.install.get_pip_args()

  if pip_args:
    args += shlex.split(pip_args)

  run_install(pip, args, upgrade, quiet, dry)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import abc
import logging
import os
import sys
from collections import Counter
from typing import Iterable, Generic, Optional, T, Type

import click
import nr.fs  # type: ignore
from databind.core import datamodel
from nr.stream import Stream  # type: ignore
from nr.utils.git import Git  # type: ignore
from termcolor import colored

from shut.changelog.manager import ChangelogManager
from shut.commands import project
from shut.model import AbstractProjectModel, Project
from shut.model.version import bump_version, parse_version, Version
from shut.renderers import get_version_refs, VersionRef
from shut.utils.io.virtual import VirtualFiles
from shut.utils.text import substitute_ranges

logger = logging.getLogger(__name__)


@datamodel
class Args:
  version: Optional[Version]
  major: bool
  minor: bool
  patch: bool
  post: bool
  snapshot: bool
  tag: bool
  push: bool
  dry: bool
  warnings_as_errors: bool
  skip_checks: bool
  skip_update: bool
  force: bool
  allow_lower: bool = False


class VersionBumpData(Generic[T], metaclass=abc.ABCMeta):

  def __init__(self, args: Args, project: Project, obj: T) -> None:
    self.args = args
    self.project = project
    self.obj = obj

  def loaded(self) -> None:
    pass

  @abc.abstractmethod
  def get_snapshot_version(self) -> Version:
    pass

  def get_version_refs(self) -> Iterable[VersionRef]:
    yield from get_version_refs(self.obj)

  def bump_to_version(self, target_version: Version) -> Iterable[str]:
    """
    Called to bump to the specified *target_version*. The default implementation uses the
    version refs provided by #get_version_refs() to bump.
    """

    version_refs = list(self.get_version_refs())

    print()
    print(f'bumping {len(version_refs)} version reference(s)')
    for filename, refs in Stream.groupby(version_refs, lambda r: r.filename, collect=list):
      with open(filename) as fp:
        content = fp.read()

      if len(refs) == 1:
        ref = refs[0]
        print(f'  {colored(nr.fs.rel(ref.filename), "cyan")}: {ref.value} → {target_version}')
      else:
        print(f'  {colored(nr.fs.rel(ref.filename), "cyan")}:')
        for ref in refs:
          print(f'    {ref.value} → {target_version}')

      content = substitute_ranges(
        content,
        ((ref.start, ref.end, target_version) for ref in refs),
      )

      if not self.args.dry:
        with open(filename, 'w') as fp:
          fp.write(content)

    changed_files = set(x.filename for x in version_refs)

    # Release the staged changelog.
    managers = list(self.get_changelog_managers())
    managers = [m for m in managers if m.unreleased.exists()]
    if managers:
      print()
      print('release staged changelog' + ('s' if len(managers) > 1 else ''))
      for manager in managers:
        changed_files.add(manager.unreleased.filename)
        if self.args.dry:
          changelog = manager.version(target_version)
        else:
          changelog = manager.release(target_version)
        changed_files.add(changelog.filename)
        print(f'  {colored(os.path.relpath(manager.unreleased.filename), "cyan")} → {os.path.relpath(changelog.filename)}')

    return changed_files

  @abc.abstractmethod
  def update(self, new_version: Version) -> VirtualFiles:
    """
    Run the "update" function for the current monorepo or package. A list of the modified files
    must be returned.
    """

  def get_changelog_managers(self) -> Iterable[ChangelogManager]:
    yield ChangelogManager(self.obj.get_changelog_directory())


def make_bump_command(
  data_class: Type[VersionBumpData[AbstractProjectModel]],
  model_type: Type[AbstractProjectModel],
) -> click.Command:

  @click.argument('version', type=parse_version, required=False)
  @click.option('--major', is_flag=True, help='bump the major number')
  @click.option('--minor', is_flag=True, help='bump the minor number')
  @click.option('--patch', is_flag=True, help='bump the patch number')
  @click.option('--post', is_flag=True, help='bump the post-release number')
  @click.option('--snapshot', is_flag=True, help='update the version number by appending the Git commit distance and shasum (note: not compatible with publishing to PyPI)')
  @click.option('--tag', is_flag=True, help='create a commit and Git tag after bumping the version')
  @click.option('--push', is_flag=True, help='push the new commit and tag to the Git "origin" remote')
  @click.option('--dry', is_flag=True, help='do not write changes to disk')
  @click.option('-w', '--warnings-as-errors', is_flag=True, help='treat check warnings as errors')
  @click.option('--skip-checks', is_flag=True, help='skip running checks before bumping')
  @click.option('--skip-update', is_flag=True, help='skip update after bumping')
  @click.option('--force', '-f', is_flag=True, help='force target version (allowing you to "bump" '
    'the same or a lower version). the flag will also result in force adding a Git tag and force '
    'pushing to the remote repository if the respective options as set (--tag and --push).')
  def bump(**kwargs):
    """
    Bump the version number to prepare a new release.
    """

    args = Args(**kwargs)
    data = data_class(args, project, project.load_or_exit(expect=model_type))
    data.loaded()
    do_bump(args, data)

  return bump


def do_bump(args: Args, data: VersionBumpData[AbstractProjectModel]) -> None:
  # Validate arguments.
  if args.push and not args.tag:
    sys.exit('error: --push can only be used with --tag')

  bump_args = 'version major minor patch post snapshot'.split()
  provided_args = {k for k in bump_args if getattr(args, k)}
  if len(provided_args) > 1:
    sys.exit('error: conflicting options: {}'.format(provided_args))
  elif not provided_args:
    print()
    print('figuring bump mode from changelog')
    entries = []
    for manager in data.get_changelog_managers():
      if manager.unreleased.exists():
        entries += manager.unreleased.entries
    if not entries:
      sys.exit('error: no changelog entries found')
    counter = Counter(entry.type_.name for entry in entries)
    bump_mode = max(entry.type_.bump_mode for entry in entries)
    assert hasattr(args, bump_mode.name), bump_mode.name
    setattr(args, bump_mode.name, True)
    print('  {} → {}'.format(
      ', '.join(f'{c} {colored(k, "cyan")}' for k, c in counter.items()),
      colored(bump_mode.name, "blue", attrs=["bold"])))

  version_refs = list(get_version_refs(data.obj))
  if not version_refs:
    sys.exit('error: no version refs found')

  # Ensure the version is the same accross all refs.
  current_version = data.obj.get_version()
  is_inconsistent = any(parse_version(x.value) != current_version for x in version_refs)
  if is_inconsistent:
    if not args.force:
      sys.exit('error: inconsistent versions across files need to be fixed first.')
    logger.warning('inconsistent versions across files were found.')

  # Bump the current version number.
  if args.post:
    new_version = bump_version(current_version, 'post')
  elif args.patch:
    new_version = bump_version(current_version, 'patch')
  elif args.minor:
    new_version = bump_version(current_version, 'minor')
  elif args.major:
    new_version = bump_version(current_version, 'major')
  elif args.snapshot:
    new_version = data.get_snapshot_version()
    if new_version < current_version:
      # The snapshot version number can be considered lower, so we'll allow it.
      args.allow_lower = True
  else:
    new_version = parse_version(args.version)

  if not new_version.pep440_compliant:
    logger.warning(f'version "{new_version}" is not PEP440 compliant.')

  # The new version cannot be lower than the current one, unless forced.
  if new_version < current_version and not (args.force or args.allow_lower):
    sys.exit(f'version "{new_version}" is lower than current version "{current_version}"')
  if str(new_version) == str(current_version) and not args.force:
    # NOTE(NiklasRosenstein): Comparing as strings to include pre-release and build number.
    logger.warning(f'new version "{new_version}" is equal to current version')
    exit(0)

  # Bump version numbers in files.
  changed_files = list(data.bump_to_version(new_version))

  if not args.skip_update:
    print()
    print('updating files')
    changed_files += data.update(new_version).abspaths('.')

  if args.tag:
    print()
    git = Git()
    files = list(git.porcelain())

    # We require that no files are currently staged.
    if any(f.mode == 'A' for f in files):
      logger.error('cannot tag with non-empty staging area')
      exit(1)

    # TODO: If this step errors (e.g. for example because the unreleased changelog file
    #   was not yet tracked by Git, making Git complain that it doesn't know about the
    #   file and it doesn't exits anymore when we add it down below), it's a bit painful
    #   to recover from it because it's a manual process of reverting the version number
    #   bumps.

    tag_name = data.obj.get_tag(new_version)
    print(f'tagging {tag_name}')

    if not args.dry:
      git.add(changed_files)
      git.commit('({}) bump version to {}'.format(data.obj.get_name(), new_version), allow_empty=True)
      git.tag(tag_name, force=args.force)

    if not args.dry and args.push:
      git.push('origin', git.get_current_branch_name(), tag_name, force=args.force)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import re
from typing import Iterable, Optional

import networkx as nx
from databind.core import datamodel, field

from .abstract import AbstractProjectModel
from .version import Version
from .requirements import VersionSelector
from .release import MonorepoReleaseConfiguration
from .package import PackageModel
from .publish import PublishConfiguration


@datamodel
class InterdependencyRef:
  filename: str
  package_name: str
  version_selector: VersionSelector
  version_start: int
  version_end: int


@datamodel
class MonorepoModel(AbstractProjectModel):

  def get_inter_dependencies(self) -> Iterable[InterdependencyRef]:
    """
    Returns an iterable that contains the names of packages in the mono repository to a list
    of their dependencies on other packages in the same repository. Note that it does so
    by regex-matching in the package configuration file rather than reading the deserialized
    package data in order to return start and end index data.
    """

    for package in self.project.packages:
      yield from self.get_inter_dependencies_for(package)

  def get_inter_dependencies_for(self, package: PackageModel) -> Iterable[InterdependencyRef]:
    """
    Like #get_inter_dependencies() but for a single package.
    """

    regex = re.compile(r'^\s*- +([A-z0-9\.\-_]+) *([^\n:]+)?$', re.M)
    package_names = set(p.name for p in self.project.packages)
    with open(package.filename) as fp:
      content = fp.read()
      for match in regex.finditer(content):
        package_name, version_selector = match.groups()
        if package_name not in package_names:
          continue
        if version_selector:
          version_selector = VersionSelector(version_selector)
        yield InterdependencyRef(package.filename, package_name, version_selector, match.start(2), match.end(2))

  def get_inter_dependencies_graph(self) -> nx.DiGraph:
    """
    Create a directed graph from the inter dependencies of packages in the mono repo.
    """

    graph = nx.DiGraph()
    for package in self.project.packages:
      graph.add_node(package.name)
    for package in self.project.packages:
      for ref in self.get_inter_dependencies_for(package):
        graph.add_edge(ref.package_name, package.name)

    return graph

  # AbstractProjectModel Overrides

  publish: PublishConfiguration = field(default_factory=PublishConfiguration)
  release: MonorepoReleaseConfiguration = field(default_factory=MonorepoReleaseConfiguration)

  def get_name(self) -> str:
    return self.name

  def get_version(self) -> Optional[Version]:
    return self.version



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import logging
import os
import sys
from typing import Iterable, List
from urllib.parse import urlparse

from shut.model import PackageModel
from shut.model.publish import WarehouseCredentials, WarehouseConfiguration
from shut.model.target import TargetId
from shut.utils.io.sp import subprocess_trimmed_call
from .core import Publisher, PublisherProvider, register_publisher_provider

logger = logging.getLogger(__name__)


def _resolve_envvars(s):
  if s and s.startswith('$'):
    value = os.getenv(s[1:])
    if not value:
      raise RuntimeError('environment variable {} is not set.'.format(s))
    return value
  return s


class WarehousePublisher(Publisher):
  """
  A publisher target that uploads files to a Warehouse (Python Package Index) using the
  "twine" command.
  """

  def __init__(self, id_: TargetId, config: WarehouseConfiguration, name: str) -> None:
    self._id = id_
    self.config = config
    self.name = name

  @classmethod
  def from_pypi_credentials(cls, id_: TargetId, creds: WarehouseCredentials) -> 'WarehousePublisher':
    config = WarehouseConfiguration(
      repository='pypi',
      test_repository='testpypi',
    ).with_creds(creds)
    return cls(id_, config, 'PyPI')

  # Publisher Overrides

  def get_description(self) -> str:
    return f'Publish the package to {self.name}.'

  def get_build_dependencies(self) -> Iterable[TargetId]:
    yield TargetId.parse('setuptools:*')

  def publish(self, files: List[str], test: bool, verbose: bool):
    command = [sys.executable, '-m', 'twine', 'upload', '--non-interactive']

    config = self.config
    repo = config.test_repository if test else config.repository
    repo_url = config.test_repository_url if test else config.repository_url
    username = config.test_username if test else config.username
    password = config.test_password if test else config.password

    username = _resolve_envvars(username)
    password = _resolve_envvars(password)

    if not repo and not repo_url:
      prefix = 'test_' if test else ''
      raise RuntimeError('missing {0}repository or {0}repository_url for PypiPublisher'.format(prefix))

    if repo:
      command += ('--repository', repo)
    if repo_url:
      command += ('--repository-url', repo_url)
    #if config.sign:
    #  command.append('--sign')
    #if config.sign_with:
    #  command += ('--sign-with', config.sign_with)
    #if config.identity:
    #  command += ('--identity', config.identity)
    if username is not None:
      command += ('--username', username)
    if password is not None:
      command += ('--password', password)
    #if config.skip_existing:
    #  command.append('--skip-existing')
    #if config.cert:
    #  command += ('--cert', config.cert)
    #if config.client_cert:
    #  command += ('--client-cert', config.client_cert)
    command += files
    command += ['--verbose']

    logger.debug('invoking twine: %s', command)
    res = subprocess_trimmed_call(command)
    return res == 0

  # Target Overrides

  @property
  def id(self) -> TargetId:
    return self._id


class WarehouseProvider(PublisherProvider[PackageModel]):

  # PublisherProvider Overrides

  def get_publishers(self, package: PackageModel) -> Iterable[Publisher]:
    publish = package.get_publish_config()
    if publish.pypi.enabled:
      yield WarehousePublisher.from_pypi_credentials(
        TargetId('warehouse', 'pypi'),
        publish.pypi.credentials)

    for name, config in publish.warehouses.items():
      display_name = name
      if config.repository_url:
        display_name = urlparse(config.repository_url).netloc
      elif config.repository:
        display_name = config.repository
      yield WarehousePublisher(TargetId('warehouse', name), config, display_name)


register_publisher_provider(PackageModel, WarehouseProvider)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from typing import Dict, Optional

from databind.core import datamodel, field


@datamodel
class WarehouseCredentials:
  username: Optional[str] = None
  password: Optional[str] = None
  test_username: Optional[str] = None
  test_password: Optional[str] = None


@datamodel
class WarehouseConfiguration(WarehouseCredentials):
  repository: Optional[str] = None
  repository_url: Optional[str] = None
  test_repository: Optional[str] = None
  test_repository_url: Optional[str] = None

  def with_creds(self, creds: WarehouseCredentials) -> 'WarehouseConfiguration':
    vars(self).update(vars(creds))
    return self


@datamodel
class PypiConfiguration:
  #: Whether publishing to PyPI is enabled.
  enabled: bool = True

  #: The credentials configuration for PyPI. Variables in the from `$VARNAME`
  #: will be substituted from environment variables.
  credentials: WarehouseCredentials = field(default_factory=WarehouseCredentials)


@datamodel
class PublishConfiguration:
  # Configuration for PyPI.
  pypi: PypiConfiguration = field(default_factory=PypiConfiguration)

  #: Additional warehouse targets to publish to.
  warehouses: Dict[str, WarehouseConfiguration] = field(default_factory=dict)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import datetime

import click

from shut.commands.commons.new import (
  GITIGNORE_TEMPLATE,
  README_TEMPLATE,
  load_author_from_git,
  get_license_file_text,
  render_template,
  write_files,
)
from shut.model import dump
from shut.model.author import Author
from shut.model.monorepo import MonorepoModel
from shut.model.release import MonorepoReleaseConfiguration
from shut.model.version import Version
from shut.utils.io.virtual import VirtualFiles
from . import mono


@mono.command(no_args_is_help=True)
@click.argument('target_directory', required=False)
@click.option('--project-name', '--name', metavar='name', required=True, help='The name of the project.')
@click.option('--author', metavar='"name <mail>"', type=Author.parse, help='The name of the author to write into the configuration file. Defaults to the name and email from the Git config.')
@click.option('--version', metavar='x.y.z', help='The version number to start counting from. Defaults to "0.0.0" (stands for "unreleased").')
@click.option('--license', metavar='name', help='The name of the license to use for the project. A LICENSE.txt file will be created.')
@click.option('--url', metavar='url', help='The URL to the project (e.g. the Git repository website).')
@click.option('--single-version', is_flag=True, help='Enable mono repository single-versioning.')
@click.option('--suffix', type=click.Choice(['yaml', 'yml']), help='The suffix for YAML files. Defaults to "yml".', default='yml')
@click.option('--dry', is_flag=True, help='Do not write files to disk.')
@click.option('-f', '--force', is_flag=True, help='Overwrite files if they already exist.')
def new(
  target_directory,
  project_name,
  author,
  version,
  license,
  url,
  single_version,
  suffix,
  dry,
  force,
):
  """
  Create files for a new Python monorepository. If the *target_directory* is specified, the files
  will be written to that directory. Otherwise the value of the --project-name argument will be
  used as the target directory.

  The following project layout will be created:

    \b
    project_name/
      .gitignore
      LICENSE.txt
      monorepo.yml
      README.md
  """

  if not target_directory:
    target_directory = project_name
  if not author:
    author = load_author_from_git() or Author('Unknown', '<unknown@example.org>')
  if not version:
    version = version or (Version('0.0.0') if single_version else None)

  package_manifest = MonorepoModel(
    name=project_name,
    version=version,
    author=author,
    license=license,
    url=url,
    release=MonorepoReleaseConfiguration(
      single_version=single_version,
    ),
  )

  template_vars = {
    'project_name': project_name,
    'version': version,
    'author': author,
    'year': datetime.date.today().year,
  }

  files = VirtualFiles()
  files.add_static('.gitignore', GITIGNORE_TEMPLATE)
  files.add_dynamic('README.md', render_template, README_TEMPLATE, template_vars)
  files.add_dynamic('monorepo.' + suffix, lambda fp: dump(package_manifest, fp))
  if license:
    files.add_dynamic('LICENSE.txt', get_license_file_text, license, template_vars)
  write_files(files, target_directory, force, dry)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from typing import Optional

import click
from termcolor import colored

from shut.commands import project
from shut.commands.commons.new import write_files
from shut.model import AbstractProjectModel
from shut.model.package import PackageModel
from shut.renderers import get_files
from shut.utils.io.virtual import VirtualFiles
from . import pkg


def verify_tag(obj: AbstractProjectModel, tag: str = None) -> bool:
  assert obj.version is not None
  expected_tag = obj.get_tag(obj.version)
  if expected_tag != tag:
    print(f'{colored("error", "red")}: tag "{tag}" does not match expected tag "{expected_tag}"')
    return False
  return True


def verify_integrity(obj: AbstractProjectModel, files: VirtualFiles) -> bool:
  modified_files = files.get_modified_files(obj.get_directory())
  if modified_files:
    print(f'{colored("error", "red")}: the following files would be modified with an update')
    for filename in modified_files:
      print(f'  {filename}')
    return False
  return True


def update_package(
  package: PackageModel,
  dry: bool = False,
  indent: int = 0,
  verify: bool = False,
  tag: Optional[str] = None,
) -> VirtualFiles:

  result = 0
  if tag and not verify_tag(package, tag):
    result = 1
  files = get_files(package)
  if verify and not verify_integrity(package, files):
    result = 1
  if result != 0:
    sys.exit(result)

  if not verify:
    write_files(files, package.get_directory(), force=True, dry=dry, indent=indent)
  return files


@pkg.command()
@click.option('--dry', is_flag=True)
@click.option('--verify', is_flag=True, help='Verify the integrity of the managed files '
  '(asserting that they would not change from running this command).')
@click.option('--verify-tag', help='Parse the version number from the specified tag and '
  'assert that it matches the version in the package configuration. (implies --verify)')
def update(dry, verify, verify_tag):
  """
  Update files auto-generated from the configuration file.
  """

  package = project.load_or_exit(expect=PackageModel)
  update_package(package, verify=verify or bool(verify_tag), tag=verify_tag)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import logging
import re
import subprocess
import sys

import click
import nr.fs  # type: ignore
import requests
from nr.utils.git import Git  # type: ignore
from termcolor import colored

from shut.model import PackageModel
from shut.model.version import parse_version
from . import project, shut

logger = logging.getLogger(__name__)


def _call(*args, **kwargs):
  proc = subprocess.Popen(*args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, **kwargs)
  stdout, _ = proc.communicate()
  if proc.returncode != 0:
    print(colored(stdout.decode(), 'red'))
    sys.exit(proc.returncode)


@shut.group()
def conda_forge():
  """
  Utility to update conda-forge recipes.
  """


@conda_forge.command()
@click.argument('package_name', required=False)
@click.argument('version', type=parse_version, required=False)
@click.option('-f', '--force', is_flag=True, help='force push to the Git repository')
def update_feedstock(package_name, version, force):
  """
  Update a conda-forge feedstock from a PyPI release. You can explicitly specify a *package_name*
  and *version*, or rely on the metadata from the package configuration file in your cwd.

  This command will clone the feedstack from the conda-forge organization on GitHub, then
  update thr version and sha256 reference and push a new branch to the repository.
  """

  if package_name or version:
    if not (package_name and version):
      sys.exit('error: need package_name AND version, OR neither')
    version = str(version)
  else:
    package = project.load_or_exit(expect=PackageModel)
    package_name, version = package.name, str(package.version)
    del package

  repo_name = 'conda-forge/{}-feedstock'.format(package_name)
  clone_url = 'git@github.com:' + repo_name
  branch_name = 'v' + version

  s_repo_name = colored(repo_name, 'cyan')
  s_package_name = colored(package_name, 'yellow')
  s_clone_url = colored(clone_url, 'cyan')
  s_branch_name = colored(branch_name, 'yellow')

  print(f'Fetching PyPI record for {s_package_name} ...')
  url = f'https://pypi.org/pypi/{package_name}/json'
  pypi_data = requests.get(url).json()
  if version not in pypi_data['releases']:
    sys.exit(f'error: no release on PyPI for {s_package_name} v{version}.')

  sdist = next((x for x in pypi_data['releases'][version] if x['packagetype'] == 'sdist'), None)
  if not sdist:
    logger.error('No sdist on PyPI for %s v%s.', s_package_name, version)
    sys.exit(1)
  sha256 = sdist['digests']['sha256']

  print(f'Updating {s_repo_name} from PyPI package {s_package_name} v{version}.')

  with nr.fs.tempdir() as tmpdir:

    print(f'Cloning {s_clone_url} ...')
    git = Git().clone(tmpdir.name, clone_url)
    git.create_branch(branch_name)

    recipe_path = nr.fs.join(tmpdir.name, 'recipe', 'meta.yaml')
    with open(recipe_path) as fp:
      recipe = fp.read()

    recipe = re.sub(r'(version = ")([^"]+)"', r'\g<1>' + version + '"', recipe)
    recipe = re.sub(r'(sha256:\s*")([^"]+)"', r'\g<1>' + sha256 + '"', recipe)

    with open(recipe_path, 'w') as fp:
      fp.write(recipe)

    print()
    subprocess.check_call(['git', '--no-pager', 'diff'], cwd=tmpdir.name)
    print()

    git.add(['.'])
    git.commit(f'Updating recipe for {package_name} v{version}.')

    print(f'pushing branch {s_branch_name} to {s_clone_url} ...')
    git.push('origin', branch_name, force=force)
